/*
 * Copyright (c) 2015-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Feb 16 2015
 */
#include "acgps.h"
#define _DEFAULT_SOURCE

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>

#ifdef HAVE_LOW_LATENCY_IOCTL
# include <sys/ioctl.h>
# include <linux/serial.h> /* for TIOCGSERIAL and ASYNC_LOW_LATENCY */
#endif

#ifdef __linux__
# include <libudev.h>
#endif

#include "codels.h"


/* --- gpstty_open --------------------------------------------------------- */

static const char *	usb_serial_to_tty(const char *serial);
static int		ftdi_sio_check_latency(int fd);

/* Open a serial port, configure to the given baud rate */
int
gpstty_open(struct gps_channel_s *chan, const char *device)
{
  const char *path;
  int fd;

  /* try to match a serial id first */
  path = usb_serial_to_tty(device);
  if (path) device = path;

  /* open non-blocking */
  fd = open(device, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd < 0) return fd;
  if (!isatty(fd)) {
    errno = ENOTTY;
    return -1;
  }

  /* discard any pending data */
  tcflush(fd, TCIOFLUSH);

#ifdef HAVE_LOW_LATENCY_IOCTL
  /* Linux: enable low latency mode
   * see /sys/bus/usb-serial/devices/.../latency_timer
   */
  {
    struct serial_struct s;
    int e;

    if (!ioctl(fd, TIOCGSERIAL, &s)) {
      if (!(s.flags & ASYNC_LOW_LATENCY)) {
        s.flags |= ASYNC_LOW_LATENCY;
        if (ioctl(fd, TIOCSSERIAL, &s)) {
          e = errno;
          switch(e) {
            case EPERM: break; /* probably not an FTDI */
            default:
              close(fd);
              errno = e;
              return -1;
          }
        }
      }
    }
  }
#endif

  if (ftdi_sio_check_latency(fd)) {
    errno = EIO;
    return -1;
  }

  /* initialize channel */
  if (chan->fd >= 0) {
    close(chan->fd);
    warnx("disconnected from %s", chan->path);
  }

  snprintf(chan->path, sizeof(chan->path), "%s", device);
  chan->baud = 0;
  chan->fd = fd;
  chan->r = chan->w = 0;
  return 0;
}


/* --- usb_serial_to_tty --------------------------------------------------- */

/* Return a tty device matching the "serial" string */

static const char *
usb_serial_to_tty(const char *serial)
{
#ifdef __linux__
  struct udev *udev;
  struct udev_enumerate *scan = NULL;
  struct udev_list_entry *ttys, *tty;
  struct udev_device *dev = NULL, *usb;
  const char *path = NULL;

  udev = udev_new();
  if (!udev) return NULL;

  /* iterate over tty devices */
  scan = udev_enumerate_new(udev);
  if (udev_enumerate_add_match_subsystem(scan, "tty")) goto done;
  if (udev_enumerate_scan_devices(scan)) goto done;

  ttys = udev_enumerate_get_list_entry(scan);
  udev_list_entry_foreach(tty, ttys) {
    const char *sysfs, *userial;

    /* get sysfs entry for the device and create a corresponding udev_device */
    if (dev) udev_device_unref(dev);
    sysfs = udev_list_entry_get_name(tty);
    dev = udev_device_new_from_syspath(udev, sysfs);

    /* get the USB device, it any */
    usb = udev_device_get_parent_with_subsystem_devtype(
      dev, "usb", "usb_device");
    if (!usb) continue;

    userial = udev_device_get_sysattr_value(usb, "serial");
    if (!userial || strcmp(userial, serial)) continue;

    /* got a match, return the tty path */
    path = strdup(udev_device_get_devnode(dev)); /* this will leak ... */
    break;
  }
  if (dev) udev_device_unref(dev);

done:
  if (scan) udev_enumerate_unref(scan);
  if (udev) udev_unref(udev);
  return path;

#else
  (void)serial; /* -Wunused-parameter */

  return NULL; /* if needed, implement this for other OSes */
#endif
}


/* --- ftdi_sio_check_latency ---------------------------------------------- */

/* Check the latency value for ftdi_sio driver */

static int
ftdi_sio_check_latency(int fd)
{
#ifdef __linux__
  struct udev *udev;
  struct udev_device *dev, *parent;
  const char *attr;
  struct stat sb;
  int s = 0;

  /* get devno */
  if (fstat(fd, &sb)) return errno;

  /* get udev_device */
  udev = udev_new();
  if (!udev) {
    warnx("cannot use udev");
    return 1;
  }

  dev = udev_device_new_from_devnum(udev, 'c', sb.st_rdev);
  if (!dev) goto done;

  /* check if it is an ftdi_sio */
  for(parent = dev; parent; parent = udev_device_get_parent(parent)) {
    attr = udev_device_get_sysattr_value(parent, "driver");
    if (attr && !strcmp(attr, "ftdi_sio")) break;
  }
  if (!parent) goto done;

  /* access latency attribute */
  attr = udev_device_get_sysattr_value(parent, "latency_timer");
  if (!attr) {
    warnx("cannot access latency_timer");
    s = 1;
    goto done;
  }

  if (atoi(attr) > 1) {
    warnx("FTDI latency timer too high: %sms, should be 1ms", attr);
    s = 1;
  }

done:
  if (dev) udev_device_unref(dev);
  udev_unref(udev);
  return s;

#else
  (void)fd; /* -Wunused-parameter */

  return 0; /* if needed, implement this for other OSes */
#endif
}


/* --- gpstty_setspeed ----------------------------------------------------- */

int
gpstty_setspeed(struct gps_channel_s *chan, uint32_t speed)
{
  struct termios t;
  speed_t baud;

  /* flush */
  chan->r = chan->w = 0;
  tcdrain(chan->fd);

  /* select baud rate */
#ifndef B9600
# define B9600 9600U
#endif
#ifndef B38400
# define B38400 38400U
#endif
#ifndef B57600
# define B57600 57600U
#endif
#ifndef B115200
# define B115200 115200U
#endif
#ifndef B230400
# define B230400 230400U
#endif
#ifndef B460800
# define B460800 460800U
#endif
#ifndef B921600
# define B921600 921600U
#endif
  switch(speed) {
    case 9600:		baud = B9600; break;
    case 38400:		baud = B38400; break;
    case 57600:		baud = B57600; break;
    case 115200:	baud = B115200; break;
    case 230400:	baud = B230400; break;
    case 460800:	baud = B460800; break;
    case 921600:	baud = B921600; break;

    default: errno = EINVAL; return -1;
  }

  /* configure line discipline */
  if (tcgetattr(chan->fd, &t)) return -1;

  t.c_iflag = IGNBRK;
  t.c_oflag = 0;
  t.c_lflag = EXTPROC;
  t.c_cflag = CS8 | CREAD | CLOCAL;
  t.c_cc[VMIN] = 0;
  t.c_cc[VTIME] = 0;

  if (cfsetspeed(&t, baud)) return -1;
  if (tcsetattr(chan->fd, TCSADRAIN, &t)) return -1;

  chan->baud = speed;
  return 0;
}


/* --- gpstty_recv --------------------------------------------------------- */

ssize_t
gpstty_recv(struct gps_channel_s *chan, const struct timeval *deadline)
{
  struct iovec iov[2];
  ssize_t s, n;

  if (chan->fd < 0) return -1;
  n = 0;

  do {
    /* feed the ring buffer */
    iov[0].iov_base = chan->buf[0] + chan->w;
    iov[1].iov_base = chan->buf[0];

    if (chan->r > chan->w) {
      iov[0].iov_len = chan->r - chan->w - 1;
      iov[1].iov_len = 0;
    } else if (chan->r > 0) {
      iov[0].iov_len = sizeof(chan->buf[0]) - chan->w;
      iov[1].iov_len = chan->r - 1;
    } else {
      iov[0].iov_len = sizeof(chan->buf[0]) - chan->w - 1;
      iov[1].iov_len = 0;
    }

    if (!iov[0].iov_len && !iov[1].iov_len) {
      /* buffer full: if some new data has been read, just return it */
      if (n) return n;

      /* otherwise, discard oldest byte */
      warnx("ring buffer full");
      chan->r = (chan->r + 1) % sizeof(chan->buf[0]);
      s = -1;
      continue;
    }

    do {
      s = readv(chan->fd, iov, 2);
    } while(s < 0 && errno == EINTR);

    if (s < 0)
      return n > 0 ? n : -1;
    else if (s == 0 && n > 0)
      return n;
    else if (s == 0) {
      struct pollfd fd = { .fd = chan->fd, .events = POLLIN };
      int timeout;

      if (deadline) {
        struct timeval tv, delay;
        gettimeofday(&tv, NULL);
        if (timercmp(deadline, &tv, <)) return 0;

        timersub(deadline, &tv, &delay);
        timeout = delay.tv_sec * 1000 + delay.tv_usec / 1000;
      } else
        timeout = 0;

      s = poll(&fd, 1, timeout);
      if (s > 0 && fd.revents & POLLHUP) { errno = ENXIO; return -1; }
    } else {
      /* mirror data (need to mirror only data before chan->r) */
      if ((size_t)s > iov[0].iov_len) {
        if (chan->r > chan->w)
          memcpy(chan->buf[1] + chan->w, iov[0].iov_base, iov[0].iov_len);
        memcpy(chan->buf[1], iov[1].iov_base, s - iov[0].iov_len);
      } else
        memcpy(chan->buf[1] + chan->w, iov[0].iov_base, s);

      chan->w = (chan->w + s) % sizeof(chan->buf[0]);
      n += s;
    }
  } while(s);

  return n;
}


/* --- gpstty_data --------------------------------------------------------- */

const void *
gpstty_data(struct gps_channel_s *chan, size_t advance, size_t *len)
{
  void *d;
  size_t s;

  /* available bytes */
  if (chan->r <= chan->w)
    s = chan->w - chan->r;
  else
    s = sizeof(chan->buf[0]) + chan->w - chan->r;

  /* output data */
  if (len) *len = s;
  d = &chan->buf[0][chan->r];

  /* advance read pointer */
  if (advance < s) s = advance;
  chan->r = (chan->r + s) % sizeof(chan->buf[0]);

  return d;
}


/* --- gpstty_sendmsg ------------------------------------------------------ */

int
gpstty_sendmsg(const struct gps_channel_s *chan, const char *fmt, ...)
{
  va_list ap;
  size_t n;

  va_start(ap, fmt);
  n = gpstty_vsendmsg(chan, fmt, ap);
  va_end(ap);

  return n;
}

int
gpstty_vsendmsg(const struct gps_channel_s *chan, const char *fmt, va_list args)
{
  char buf[256];
  size_t n;

  n = vsnprintf(buf, sizeof(buf), fmt, args);
  if (n >= sizeof(buf)) return E2BIG;

  return gpstty_send(chan, buf, n);
}


/* --- gpstty_send --------------------------------------------------------- */

int
gpstty_send(const struct gps_channel_s *chan, const void *buf, size_t n)
{
  ssize_t s;

  while (n > 0) {
    do {
      s = write(chan->fd, buf, n);
    } while (s < 0 && errno == EINTR);
    if (s < 0) return errno;

    buf += s;
    n -= s;
  }

  tcdrain(chan->fd);
  return 0;
}


/* --- gpstty_close -------------------------------------------------------- */

void
gpstty_close(struct gps_channel_s *chan)
{
  if (chan->fd >= 0) {
    close(chan->fd);
    if (*chan->path)
      warnx("disconnected from %s", chan->path);
  }

  *chan->path = 0;
  chan->baud = 0;
  chan->fd = -1;
  chan->r = chan->w = 0;
}
